from fireo.models.model import Model

from fireo.fields import Field
from fireo.fields import IDField
from fireo.fields import BooleanField
from fireo.fields import DateTime
from fireo.fields import GeoPoint
from fireo.fields import ListField
from fireo.fields import MapField
from fireo.fields import NestedModel
from fireo.fields import NumberField
from fireo.fields import ReferenceField
from fireo.fields import TextField



from fireo.fields.base_field import Field


class IDField(Field):
    """Specify custom id for models

    User can specify model id and will save with the same id in firestore otherwise it will
    return None and generate later from firestore and attached to model

    Example
    --------
    .. code-block:: python
        class User(Mode):
            user_id = IDField()

        u = User()
        u.user_id = "custom_doc_id"
        u.save()

        # After save id will be saved in `user_id`
        print(self.user_id)  # custom_doc_id
    """
    def contribute_to_model(self, model_cls, name):
        self.name = name
        setattr(model_cls, name, None)
        model_cls._meta.add_model_id(self)


from fireo.fields import errors
from fireo.fields.base_field import Field


class TextField(Field):
    """Text field for Models

    Define text for models

    allowed_attributes = ['max_length']



    Examples
    --------
        class User(Model):
            age = TextField()
    """

    allowed_attributes = ['max_length']

    def attr_max_length(self, attr_val, field_val):
        """Method for attribute max_length"""
        return field_val[:attr_val]

    # override method
    def db_value(self, val):
        if type(val) is str or val is None:
            # check if user defined to set the value as lower case
            if self.model_cls._meta.to_lowercase:
                return val.lower()
            return val
        raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {str}, '
                                      f'got {type(val)}')



from fireo.database.errors import DBConnectionError
from google.cloud import firestore


class Database:
    """Create connection with google cloud firestore

    Responsible to create and verify connection is successfully establish.

    If you’re running in Compute Engine or App Engine, authentication should “just work”.
    If you’re developing locally, the easiest way to authenticate is using the Google Cloud SDK:

    If you’re running your application elsewhere, you should download a service account JSON keyfile
    and point to it using an environment variable

    More about Authentication: https://googleapis.dev/python/google-api-core/latest/auth.html

    Attributes
    ----------
    conn : firestore.Client()
        return the connection from firestore

    Methods
    -------
    connect():
        create a connection with firestore

    Raises
    ------
    DBConnectionError:
        Missing or wrong project id or credentials
    """

    def __init__(self):
        self._conn = None

    def connect(self,credentials=None, from_file=None):
        try:
            if credentials:
                self._conn = firestore.Client(credentials=credentials)
            elif from_file:
                self._conn = firestore.Client.from_service_account_json(from_file)
            else:
                raise DBConnectionError("Credentials or service account json file required to connect with firestore")
        except Exception as e:
            raise DBConnectionError(e) from e

    @property
    def conn(self):
        if self._conn is None:
            self._conn = firestore.Client()
        return self._conn



import inspect

from fireo.database import db
from fireo.utils import utils
from fireo.queries import errors


class BaseQuery:
    """Create connection with firestore and provide ref for model collection

    Methods
    -------
    set_collection_path(path, key):
        Set collection path

    get_ref():
        Provide firestore ref for model collection

    set_group_collection():
        setter method for group collection

    validate_key():
        Validate the key
    """

    def __init__(self, model_cls):
        self.model_cls = model_cls
        # if collection path and model key both are not available then
        # collection_name will be the base collection path
        self.collection_path = model_cls.collection_name
        # Firestore allow to get documents from group collection
        self.group_collection = False

    def set_collection_path(self, path=None, key=None):
        """Set collection path"""
        # Check collection path is given if not then get it from model key
        if path:
            self.collection_path = path + '/' + self.model_cls.collection_name
        elif key:
            self.collection_path = utils.collection_path(key)

    def get_ref(self):
        """Provide firestore ref for model collection"""

        #  Validate the key
        self.validate_key()

        if self.group_collection:
            return db.conn.collection_group(self.collection_path)

        return db.conn.collection(self.collection_path)

    def set_group_collection(self, is_group):
        """Set group collection"""
        self.group_collection = is_group

    def validate_key(self):
        """Validate key, Key collection must be the same as the model collection name

        Examples
        --------
        .. code-block:: python
            class User(Model):
                name = TextField()

            u = User.collection.create(name="Azeem")

            Student.collection.get(u.key)  # Invalid key because collection is student and key is for User

            # Valid key
            User.collection.get(u.key)

        Raises
        ------
        InvalidKey:
            If collection path in key is not same as the model collection name
        """
        # Get model name because when document is saving model instance is passed instead of
        # model class for modifying the model instance
        if inspect.isclass(self.model_cls):
            model_name = self.model_cls.__name__
        else:
            model_name = self.model_cls.__class__.__name__
        # Check collection is not parent
        if '/' not in self.collection_path:
            # check collection path is the same as collection name
            if self.collection_path != self.model_cls.collection_name:
                raise errors.InvalidKey(f'Invalid key is given, expected "{model_name}" type key, '
                                        f'got "{self.collection_path}" type key')
        else:
            # If collection path is contains the parent then
            # Get last name from collection path
            # For example if collection path is "user/doc_id/student/student_doc_id then student
            # will be the collection name
            key_collection = self.collection_path.split('/')[-1]
            if key_collection != self.model_cls.collection_name:
                raise errors.InvalidKey(f'Invalid key is given, expected "{model_name}" type key, '
                                        f'got "{key_collection}" type key')



from google.cloud import firestore

from fireo.database import db


class Transaction:
    """Provide a callable object to use as a transactional decorater."""

    def __init__(self, to_wrap):
        self.to_wrap = to_wrap

    def __call__(self, transaction, *args, **kwargs):
        @firestore.transactional
        def operate_transaction(transaction, *args, **kwargs):
            return self.to_wrap(transaction, *args, **kwargs)

        return operate_transaction(transaction, *args, **kwargs)


from fireo.fields import ReferenceField, NestedModel
from fireo.queries import errors
from fireo.utils import utils
from google.cloud import firestore


class ModelWrapper:
    """Convert query result into Model instance"""
    @classmethod
    def from_query_result(cls, model, doc, nested_doc=False):
        parent_key = None
        if nested_doc:
            doc_dict = doc
        elif doc:
            parent_key = utils.get_parent_doc(doc.reference.path)
            if doc.to_dict():
                doc_dict = doc.to_dict()
            else:
                return None
        else:
            return None

        # instance values is changed according to firestore
        # so mark it modified this will help later for figuring
        # out the updated fields when need to update this document
        setattr(model, '_instance_modified', True)
        for k, v in doc_dict.items():
            field = model._meta.get_field_by_column_name(k)
            # if missing field setting is set to "ignore" then
            # get_field_by_column_name return None So, just skip this field
            if field is None:
                continue
            # Check if it is Reference field
            if isinstance(field, ReferenceField):
                val = ReferenceFieldWrapper.from_doc_ref(model, field, field.field_value(v))
            elif isinstance(field, NestedModel):
                nested_doc_val = field.field_value(v)
                if nested_doc_val:
                    val = NestedModelWrapper.from_model_dict(field, nested_doc_val)
                else:
                    val = None
            else:
                val = field.field_value(v)
            setattr(model, field.name, val)

        # If parent key is None but here is parent key from doc then set the parent for this model
        # This is case when you filter the documents parent key not auto set just set it
        if not model.parent and parent_key is not None:
            model.parent = parent_key

        # If it is not nested model then set the id for this model
        if not nested_doc:
            setattr(model, '_id', doc.id)
            # save the firestore reference doc so that further actions can be performed (i.e. collections())
            model._meta.set_reference_doc(doc.reference)
            # even though doc.reference currently points to self, there is no guarantee this will be true
            # in the future, therefore we should store the create time and update time separate.
            model._meta._firestore_create_time = doc.create_time
            model._meta._firestore_update_time = doc.update_time
        return model


class NestedModelWrapper:
    """Get nested document"""
    @classmethod
    def from_model_dict(cls, field, doc):
        model = field.nested_model()
        return ModelWrapper.from_query_result(model, doc, nested_doc=True)


class ReferenceFieldWrapper:
    """Get reference documents

    If auto_load is True then load the document otherwise return `ReferenceDocLoader` object and later user can use
    `get()` method to retrieve the document
    """
    @classmethod
    def from_doc_ref(cls, parent_model, field, ref):
        if not ref:
            return None

        ref_doc = ReferenceDocLoader(parent_model, field, ref)

        if field.auto_load:
            return ref_doc.get()
        return ref_doc


class ReferenceDocLoader:
    """Get reference doc and Convert into model instance"""
    def __init__(self, parent_model, field, ref):
        self.parent_model = parent_model
        self.field = field
        self.ref = ref

    def get(self):
        doc = self.ref.get()
        if not doc.exists:
            raise errors.ReferenceDocNotExist(f'{self.field.model_ref.collection_name}/{self.ref.id} not exist')
        model = ModelWrapper.from_query_result(self.field.model_ref(), doc)

        # if on_load method is defined then call it
        if self.field.on_load:
            method_name = self.field.on_load
            getattr(self.parent_model, method_name)(model)
        return model



from fireo.database import db
from fireo.fields import errors
from fireo.fields.base_field import Field
from fireo.utils import utils
from google.cloud import firestore


class ReferenceField(Field):
    """Reference of other model

    A DocumentReference refers to a document location in a Firestore database and
    can be used to write, read, or listen to the location. The document at the referenced
    location may or may not exist.

    Attributes
    ----------
    allowed_attribute: ['auto_load']
        Allow reference field to load automatically or not

    model_ref:
        Reference of the model

    auto_load:
        Reference field load setting, load it auto or not

    on_load:
        Call user specify method when reference document load

    Methods
    -------
    attr_auto_load():
        Method for attribute auto_load

    attr_on_load():
        Method for attribute on_load

    Raises
    ------
    AttributeTypeError:
        If given type is not supported by attribute
    """

    allowed_attributes = ['auto_load', 'on_load']

    def __init__(self, model_ref, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Check model ref class is subclass for Model
        from fireo.models import Model
        if not issubclass(model_ref, Model):
            raise errors.ReferenceTypeError(f'Reference model "{model_ref.__name__}" must be inherit from Model class')
        self.model_ref = model_ref
        self.auto_load = True
        self.on_load = None

    # Override method
    def field_value(self, val):
        v = self.field_attribute.parse(val)
        return v

    # Override method
    def db_value(self, model):
        # if no model is provided then return None
        if model is None:
            return None
        # check reference model and passing model is same
        if not issubclass(model.__class__, self.model_ref):
            raise errors.ReferenceTypeError(f'Invalid reference type. Field "{self.name}" required value type '
                                            f'"{self.model_ref.__name__}", but got "{model.__class__.__name__}"')
        # Get document reference from firestore
        return firestore.DocumentReference(*utils.ref_path(model.key), client=db.conn)

    def attr_auto_load(self, attr_val, field_val):
        """Attribute method for auto load

        Allow users to load reference documents automatically or just return the reference
        and user itself call the `get()` method to load the document.
        """
        if type(attr_val) is not bool:
            raise errors.AttributeTypeError(f'Attribute auto_load only accept bool type, got {type(attr_val)} in '
                                            f'model "{self.model_cls.__name__}" field "{self.name}"')
        self.auto_load = attr_val
        return field_val

    def attr_on_load(self, method_name, field_val):
        """Attribute method for on load

        Call user specify method when reference document is loaded
        """
        try:
            m = getattr(self.model_cls, method_name)
            if not callable(m):
                raise errors.AttributeTypeError(f'Attribute {m} is not callable in model "{self.model_cls.__name__}" '
                                                f'field "{self.name}"')
            self.on_load = method_name
        except AttributeError as e:
            raise errors.AttributeMethodNotDefined(f'Method "{method_name}" is not defined for attribute on_load in '
                                                   f'model "{self.model_cls.__name__}" field "{self.name}"') from e

        return field_val



from google.cloud import firestore

from fireo.utils.utils import GeoPoint
from fireo.database import connection, db
from fireo.transaction import Transaction


def transaction(**kwargs):
    return db.conn.transaction(**kwargs)


def batch():
    return db.conn.batch()


def transactional(to_wrap):
    return Transaction(to_wrap)


def ListUnion(values):
    return firestore.ArrayUnion(values)


def ListRemove(values):
    return firestore.ArrayRemove(values)


def Increment(value):
    return firestore.Increment(value)

def list_collections():
    """Returns a list of all collections"""
    return [ c.id for c in db.conn.collections() ]


from datetime import datetime

from fireo.fields import Field, errors
from google.cloud import firestore
from google.cloud.firestore_v1.transforms import Sentinel
from google.api_core.datetime_helpers import DatetimeWithNanoseconds


class DateTime(Field):
    """Date Time field for firestore

    allowed_attributes = [auto]

    Examples
    --------
        class User(Model):
            created = DateTime()

        u = User()
        u.created = datetime.datetime.now()
    """

    allowed_attributes = ['auto']

    def attr_auto(self, attr_val, field_val):
        """Method for attribute auto"""
        if field_val is None and attr_val:
            return firestore.SERVER_TIMESTAMP
        return field_val

    # Override method
    def db_value(self, val):
        if any([
            isinstance(val, DatetimeWithNanoseconds),
            isinstance(val, datetime),
            isinstance(val, Sentinel),
            isinstance(val, type(None)),
        ]):
            return val
        raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {datetime}, '
                                      f'got {type(val)}')



from fireo.queries import query_wrapper
from fireo.queries.base_query import BaseQuery
from fireo.utils import utils


class GetQuery(BaseQuery):
    """Get model from firestore

    Get model from firestore using Managers

    Examples
    --------
    .. code-block:: python
        class User(Mode):
            name = TextField()
            age = NumberField()

        u = User(name="Azeem", age=26)
        u.save()

        # Getting model
        user = User.collection.get(u.key)
        print(user.name)  # Azeem
        print(user.age)  # 26

    Methods
    -------
    _raw_exec():
        private method to get documents from firestore

    exec():
        Get document from firestore and wrap them into model
    """

    def __init__(self, model_cls, key):
        super().__init__(model_cls)
        super().set_collection_path(key=key)
        self.model = model_cls()
        # set parent to this model if any
        self.model.parent = utils.get_parent_doc(key)
        # Attach key to this model for updating this model
        # Purpose of attaching this key is user can update
        # this model after getting it
        #
        # For example:
        #   u = User.collection.get(user_key)
        #   u.name = "Updated Name"
        #   u.update()
        self.model._update_doc = key
        self.id = utils.get_id(key)

    def _raw_exec(self, transaction=None):
        """Get firestore reference and then get document based on id"""
        ref = self.get_ref().document(self.id)
        return ref.get(transaction=transaction)

    def exec(self, transaction=None):
        """Wrap the query result into model instance"""
        return query_wrapper.ModelWrapper.from_query_result(self.model, self._raw_exec(transaction))



from google.cloud.firestore_v1 import ArrayUnion, ArrayRemove

from fireo.fields import Field, errors


class ListField(Field):
    """Array field for firestore

    Example
    -------
        class User(Model):
            subjects = ListField()

        u = User()
        u.subjects = ['English', 'Math']
    """

    # Override method
    def db_value(self, val):
        if type(val) in [list, ArrayUnion, ArrayRemove] or val is None:
            # check if user defined to set the value as lower case
            if self.model_cls._meta.to_lowercase:
                return [v.lower() if type(v) is str else v for v in val]
            return val
        raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {list}, '
                                      f'got {type(val)}')



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='fireo',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.1.5',

    description="FireO ORM is specifically designed for the Google's Firestore",
    long_description=long_description,
    long_description_content_type="text/markdown",

    # The project's main homepage.
    url="https://github.com/octabytes/FireO",

    # Author details
    author="OctaByte",
    author_email="Dev@octabyte.io",

    # Choose your license
    license="Apache 2.0",

    # See https://pypi.org/classifiers/
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",

        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ],

    # What does your project relate to?
    keywords='Python Firestore Models ORM Google Cloud Firebase',

    # Add all packages except tests
    packages=find_packages('src', exclude=['tests']),

    package_dir={'': 'src'},

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['google-cloud-firestore'],
)



"""Field related errors"""


class NumberRangeError(Exception):
    pass


class UnSupportedAttribute(Exception):
    pass


class InvalidFieldType(Exception):
    pass


class AttributeTypeError(Exception):
    pass


class NestedModelTypeError(Exception):
    pass


class ReferenceTypeError(Exception):
    pass


class MissingFieldOptionError(Exception):
    pass


class FieldNotFound(Exception):
    pass


class RequiredField(Exception):
    pass


class FieldValidationFailed(Exception):
    pass


class ValidatorNotCallable(Exception):
    pass


class AttributeMethodNotDefined(Exception):
    pass




from google.cloud.firestore_v1 import Increment

from fireo.fields import errors
from fireo.fields.base_field import Field


class NumberField(Field):
    """Number field for Models

    Define numbers for models integer, float etc

    allowed_attributes = ['int_only', 'float_only', range]

    Examples
    --------
        class User(Model):
            age = NumberField()
    """

    allowed_attributes = ['int_only', 'float_only', 'range']

    def attr_range(self, attr_val, field_val):
        """Method for attribute range"""
        try:
            start, stop = attr_val
        except TypeError:
            start = attr_val
            stop = None

        if start and field_val < start:
            raise errors.NumberRangeError(f'Field "{self.name}" expect number must be grater or equal '
                                          f'than {start}, given {field_val}')
        if stop and field_val > stop:
            raise errors.NumberRangeError(f'Field "{self.name}" expect number must be less than or equal '
                                          f'to {stop}, given {field_val}')

        return field_val

    def attr_int_only(self, attr_val, field_val):
        """Method for attribute int_only"""
        if attr_val and type(field_val) is not int:
            raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {int} type, '
                                          f'got {type(field_val)}')
        return field_val

    def attr_float_only(self, attr_val, field_val):
        """Method for attribute float_only"""
        if attr_val and type(field_val) is not float:
            raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {float} type, '
                                          f'got {type(field_val)}')
        return field_val

    # override method
    def db_value(self, val):
        if type(val) in [int, float, Increment] or val is None:
            return val
        raise errors.InvalidFieldType(f'Invalid field type. Field "{self.name}" expected {int} or {float}, '
                                      f'got {type(val)}')


from fireo.fields.base_field import Field
from fireo.fields.id_field import IDField
from fireo.fields.boolean_field import BooleanField
from fireo.fields.data_time import DateTime
from fireo.fields.geo_point import GeoPoint
from fireo.fields.list_field import ListField
from fireo.fields.map_field import MapField
from fireo.fields.nested_model import NestedModel
from fireo.fields.number_field import NumberField
from fireo.fields.reference_field import ReferenceField
from fireo.fields.text_field import TextField



import base64
import itertools
from fireo.queries import query_wrapper


class QueryIterator:
    """Iterator the documents coming from Firestore

    Iterate each document that is coming from Firestore and wrap it to the model
    QueryIterator also contain the other information for example Firestore query, cursor etc

    Methods
    -------
    fetch_next():
        Fetch next results
    """

    def __init__(self, query):
        self.query = query
        self.model_cls = query.model.__class__
        self.docs = query.query().stream(query.query_transaction)
        # Get offset for next fetch
        self.offset = query.n_limit
        query.cursor_dict['offset'] = query.n_limit
        # Hold the last doc for next fetch
        self.last_doc = None
        self.last_doc_key = None
        self.fetch_end = False

    def __iter__(self):
        return self

    def __next__(self):
        try:
            doc = next(self.docs, None)
            if doc:
                # Suppose this is the last doc
                self.last_doc = doc
                m = query_wrapper.ModelWrapper.from_query_result(self.model_cls(), doc)
                m._update_doc = self.query._update_doc_key(m)
                # Suppose this is last doc
                self.last_doc_key = m.key
                return m
            self.fetch_end = True
            # Save last doc key in cursor
            self.query.cursor_dict['last_doc_key'] = self.last_doc_key
            raise StopIteration
        except Exception:
            raise StopIteration

    def next_fetch(self, limit=None):
        """Fetch next results"""

        # Make sure limit is apply in previous query otherwise there is no need to fetch next
        if self.offset:
            # check if fetch end then use last doc otherwise use the offset
            if self.fetch_end:
                self.fetch_end = False
                q = self.query.query().start_after(self.last_doc)
            else:
                q = self.query.query().offset(self.offset)

            # Apply new Limit if there is any
            if limit:
                q = q.limit(limit)
                self.offset += limit
            else:
                self.offset += self.query.n_limit

            # Update offset in cursor
            self.query.cursor_dict['offset'] = self.offset

            self.docs = itertools.chain(self.docs, q.stream())

    @property
    def cursor(self):
        return base64.b64encode(bytes(str(self.query.cursor_dict), 'utf-8'))



from fireo.fields import errors
from fireo.fields.base_field import Field


class NestedModel(Field):
    """Model inside another model"""

    def __init__(self, model, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Check nested model class is subclass for Model
        from fireo.models import Model
        if not issubclass(model, Model):
            raise errors.NestedModelTypeError(f'Nested model "{model.__name__}" must be inherit from Model class')
        self.nested_model = model

    def valid_model(self, model_instance):
        """Check nested model and passing model is same"""

        # return False if no Nested model apply
        if model_instance is None:
            return False

        if self.nested_model == model_instance.__class__:
            return True
        raise errors.NestedModelTypeError(f'Invalid nested model type. Field "{self.name}" required value type '
                                          f'"{self.nested_model.__name__}", but got '
                                          f'"{model_instance.__class__.__name__}"')

